import os
import sys
import datetime
import importlib
import inspect
import json
import functools
import platform
import re
import time
import warnings
from packaging import version as version_mod
from pathlib import Path
import pkgutil
import traceback
import platform
from typing import Union, List

import numpy as np
from qtpy import QtCore
from qtpy.QtCore import QLocale

from pymodaq.utils import logger as logger_module
from pymodaq.utils.config import get_set_preset_path, Config
from pymodaq.utils.messenger import deprecation_msg
from pymodaq.utils.qvariant import QVariant

if version_mod.parse(platform.python_version()) >= version_mod.parse('3.8'):  # from version 3.8 this feature is included in the
    # standard lib
    from importlib import metadata
else:
    import importlib_metadata as metadata

if version_mod.parse(platform.python_version()) >= version_mod.parse('3.9'):
    # from version 3.9 the cache decorator is available
    from functools import cache
else:
    from functools import lru_cache as cache


if version_mod.parse(platform.python_version()) >= version_mod.parse('3.9'):
    # from version 3.9 the cache decorator is available
    from functools import cache
else:
    from functools import lru_cache as cache

logger = logger_module.set_logger(logger_module.get_module_name(__file__))

plot_colors = [(255, 255, 255), (255, 0, 0), (0, 255, 0), (0, 0, 255), (14, 207, 189), (207, 14, 166), (207, 204, 14)]
config = Config()


def __getattr__(name):
    if name in ['Axis', 'NavAxis', 'ScaledAxis', 'ScalingOptions', 'Data', 'DataTimeStamped', 'DataFromPlugins',
                'DataToEmit', 'DataToExport']:
        data_mod = importlib.import_module('.data', 'pymodaq.utils')
        deprecation_msg('Loading Axis or Data and their derived classes from daq_utils is deprecated, import them from'
                        ' pymodaq.utils.data module', 3)
        return getattr(data_mod, name)
    else:
        raise AttributeError


def load_config():
    deprecation_msg(f'config methods must now be  imported from the pymodaq.utils.messenger.cnfig module')
    return Config()


def set_logger(*args, **kwargs):
    deprecation_msg(f'Logger methods must now be  imported from the pymodaq.utils.logger module', 3)
    return logger_module.set_logger(*args, **kwargs)


def get_module_name(*args, **kwargs):
    deprecation_msg(f'Logger methods must now be  imported from the pymodaq.utils.logger module', 3)
    return logger_module.get_module_name(*args, **kwargs)


def is_64bits():
    return sys.maxsize > 2**32


def timer(func):
    """Print the runtime of the decorated function"""
    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        start_time = time.perf_counter()    # 1
        value = func(*args, **kwargs)
        end_time = time.perf_counter()      # 2
        run_time = end_time - start_time    # 3
        print(f"Finished {func.__name__!r} in {run_time:.4f} secs")
        return value
    return wrapper_timer


def get_version():
    """Obtain pymodaq version from the VERSION file

    Follows the layout from the packaging tool hatch, hatchling
    """
    DEFAULT_PATTERN = r'(?i)^(__version__|VERSION) *= *([\'"])v?(?P<version>.+?)\2'

    with open(str(Path(__file__).parent.parent.joinpath('resources/VERSION')), 'r') as fvers:
        contents = fvers.read().strip()
        match = re.search(DEFAULT_PATTERN, contents, flags=re.MULTILINE)
        groups = match.groupdict()
        if 'version' not in groups:
            raise ValueError('no group named `version` was defined in the pattern')
    return groups['version']


def copy_preset():                          # pragma: no cover
    path = get_set_preset_path().joinpath('preset_default.xml')
    if not path.exists():  # copy the preset_default from pymodaq folder and create one in pymodad's local folder
        with open(str(Path(__file__).parent.parent.joinpath('resources/preset_default.xml')), 'r') as file:
            path.write_text(file.read())


def set_qt_backend():
    backend_present = True
    if config('qtbackend', 'backend').lower() not in [mod.lower() for mod in sys.modules]:
        backend_present = False
        logger.warning(f"The chosen Qt backend ({config('qtbackend', 'backend')}) has not been installed...\n"
                       f"Trying another...")
        backends = config('qtbackend', 'backends')
        backends.pop(config('qtbackend', 'backend'))
        for backend in backends:
            if backend.lower() in [mod.lower() for mod in sys.modules]:
                backend_present = True
                break

    if backend_present:
        os.environ['QT_API'] = config('qtbackend', 'backend')
        logger.info('************************')
        logger.info(f"{config('qtbackend', 'backend')} Qt backend loaded")
        logger.info('************************')
    else:
        msg = f"No Qt backend could be found in your system, please install either pyqt5/6 or pyside2/6." \
              f"pyqt5 is still preferred, while pyqt6 should mostly work."
        logger.critical(msg)
        warnings.warn(msg, FutureWarning, 3)
        print(msg.upper())


class JsonConverter:
    def __init__(self):
        super().__init__()

    @classmethod
    def trusted_types(cls):
        return ['float', 'int', 'str', 'datetime', 'date', 'time', 'tuple', 'list', 'bool', 'bytes',
                'float64']

    @classmethod
    def istrusted(cls, type_name):
        return type_name in cls.trusted_types()

    @classmethod
    def object2json(cls, obj):
        dic = dict(module=type(obj).__module__, type=type(obj).__name__, data=repr(obj))
        return json.dumps(dic)

    @classmethod
    def json2object(cls, jsonstring):
        try:
            dic = json.loads(jsonstring)
            if isinstance(dic, dict):
                if dic['type'] in cls.trusted_types():
                    return eval(dic['data'])
                else:
                    return dic
            else:                                               # pragma: no cover
                return dic
        except Exception:
            return jsonstring


def decode_data(encoded_data):
    """
    Decode QbyteArrayData generated when drop items in table/tree/list view
    Parameters
    ----------
    encoded_data: QByteArray
                    Encoded data of the mime data to be dropped
    Returns
    -------
    data: list
            list of dict whose key is the QtRole in the Model, and the value a QVariant

    """
    data = []

    ds = QtCore.QDataStream(encoded_data, QtCore.QIODevice.ReadOnly)
    while not ds.atEnd():
        row = ds.readInt32()
        col = ds.readInt32()

        map_items = ds.readInt32()
        item = {}
        for ind in range(map_items):
            key = ds.readInt32()
            #TODO check this is fine
            value = QVariant()
            #value = None
            ds >> value
            item[QtCore.Qt.ItemDataRole(key)] = value.value()
        data.append(item)
    return data


#############################

def capitalize(string, Nfirst=1):
    """
    Returns same string but with first Nfirst letters upper
    Parameters
    ----------
    string: (str)
    Nfirst: (int)
    Returns
    -------
    str
    """
    return string[:Nfirst].upper() + string[Nfirst:]


def uncapitalize(string, Nfirst=1):
    return string[:Nfirst].lower() + string[Nfirst:]


def getLineInfo():
    """get information about where the Exception has been triggered"""
    tb = sys.exc_info()[2]
    res = ''
    for t in traceback.format_tb(tb):
        res += t
    return res


class ThreadCommand:
    """Generic object to pass info (command) and data (attribute) between thread or objects using signals

    Parameters
    ----------
    command: str
        The command to be analysed for further action
    attribute: any type
        the attribute related to the command. The actual type and value depend on the command and the situation
    attributes: deprecated, attribute should be used instead

    Attributes
    ----------
    command : str
        The command to be analysed for further action
    attribute : any type
        the attribute related to the command. The actual type and value depend on the command and the situation
    """

    def __init__(self, command: str, attribute=None, attributes=None):
        if not isinstance(command, str):
            raise TypeError(f'The command in a Threadcommand object should be a string, not a {type(command)}')
        self.command = command
        if attribute is None and attributes is not None:
            deprecation_msg('ThreadCommand signature changed, use attribute in place of attribute')
            self.attribute = attributes
            self.attributes = attributes
        self.attribute = attribute

    def __repr__(self):
        return f'Threadcommand: {self.command} with attribute {self.attribute}'


def ensure_ndarray(data):
    """
    Make sure data is returned as a numpy array
    Parameters
    ----------
    data

    Returns
    -------
    ndarray
    """
    if not isinstance(data, np.ndarray):
        if isinstance(data, list):
            data = np.array(data)
        else:
            data = np.array([data])
    return data


def setLocale():
    """
    defines the Locale to use to convert numbers to strings representation using language/country conventions
    Default is English and US
    """
    language = getattr(QLocale, config('style', 'language'))
    country = getattr(QLocale, config('style', 'country'))
    QLocale.setDefault(QLocale(language, country))


def recursive_find_files_extension(ini_path, ext, paths=[]):
    with os.scandir(ini_path) as it:
        for entry in it:
            if os.path.splitext(entry.name)[1][1:] == ext and entry.is_file():
                paths.append(entry.path)
            elif entry.is_dir():
                recursive_find_files_extension(entry.path, ext, paths)
    return paths


def recursive_find_files(ini_path, exp='make_enum', paths=[],
                         filters=['build']):
    for child in Path(ini_path).iterdir():
        if child.is_dir():
            recursive_find_files(child, exp, paths, filters)
        else:
            if exp in child.stem:
                if not any([filt in str(child) for filt in filters]):
                    paths.append(child)
    return paths


def recursive_find_expr_in_files(ini_path, exp='make_enum', paths=[],
                                 filters=['.git', '.idea', '__pycache__', 'build', 'egg', 'documentation', '.tox'],
                                 replace=False, replace_str=''):

    for child in Path(ini_path).iterdir():
        if not any(filt in str(child) for filt in filters):
            if child.is_dir():
                recursive_find_expr_in_files(child, exp, paths, filters, replace=replace, replace_str=replace_str)
            else:
                try:
                    found = False
                    with child.open('r') as f:
                        replacement = ''
                        for ind, line in enumerate(f):
                            if exp in line:
                                found = True
                                paths.append([child, ind, line])
                                if replace:
                                    replacement += line.replace(exp, replace_str)
                            else:
                                if replace:
                                    replacement += line
                    if replace and found:
                        with child.open('w') as f:
                            f.write(replacement)
                except Exception:
                    pass
    return paths


def count_lines(ini_path, count=0, filters=['lextab', 'yacctab','pycache', 'pyc']):
    # if Path(ini_path).is_file():
    #     with Path(ini_path).open('r') as f:
    #         count += len(f.readlines())
    #     return count
    for child in Path(ini_path).iterdir():
        if child.is_dir():
            count = count_lines(child, count)
        else:
            try:
                if not any([filt in child.name for filt in filters]):
                    if '.py' in child.name:
                        with child.open('r') as f:
                            count += len(f.readlines())
                else:
                    print(child.stem)
            except Exception:
                pass
    return count


def remove_spaces(string):
    """
    return a string without any white spaces in it
    Parameters
    ----------
    string

    Returns
    -------

    """
    return ''.join(string.split())


def rint(x):
    """
    almost same as numpy rint function but return an integer
    Parameters
    ----------
    x: (float or integer)

    Returns
    -------
    nearest integer
    """
    return int(np.rint(x))


def elt_as_first_element(elt_list, match_word='Mock'):
    if not hasattr(elt_list, '__iter__'):
        raise TypeError('elt_list must be an iterable')
    if elt_list:
        ind_elt = 0
        for ind, elt in enumerate(elt_list):
            if not isinstance(elt, str):
                raise TypeError('elt_list must be a list of str')
            if match_word in elt:
                ind_elt = ind
                break
        plugin_match = elt_list[ind_elt]
        elt_list.remove(plugin_match)
        plugins = [plugin_match]
        plugins.extend(elt_list)
    else:
        plugins = []
    return plugins


def elt_as_first_element_dicts(elt_list, match_word='Mock', key='name'):
    if not hasattr(elt_list, '__iter__'):
        raise TypeError('elt_list must be an iterable')
    if elt_list:
        ind_elt = 0
        for ind, elt in enumerate(elt_list):
            if not isinstance(elt, dict):
                raise TypeError('elt_list must be a list of dicts')
            if match_word in elt[key]:
                ind_elt = ind
                break
        plugin_match = elt_list[ind_elt]
        elt_list.remove(plugin_match)
        plugins = [plugin_match]
        plugins.extend(elt_list)
    else:
        plugins = []
    return plugins


def find_keys_from_val(dict_tmp: dict, val: object):
    """Returns the keys from a dict if its value is matching val"""
    return [k for k, v in dict_tmp.items() if v == val]


def find_object_if_matched_attr_name_val(obj, attr_name, attr_value):
    """check if an attribute  key/value pair match in a given object

    Parameters
    ----------
    obj: object
    attr_name: str
        attribute name to look for in the object
    attr_value: object
        value to match

    Returns
    -------
    bool: True if the key/value pair has been found in dict_tmp

    """
    if hasattr(obj, attr_name):
        if getattr(obj, attr_name) == attr_value:
            return True
    return False


def find_objects_in_list_from_attr_name_val(objects: List[object], attr_name: str,
                                            attr_value: object, return_first=True):
    """ lookup within a list of objects. Look for the objects within the list which has the correct attribute name,
    value pair

    Parameters
    ----------
    objects: list
        list of objects
    attr_name: str
        attribute name to look for in the object
    attr_value: object
        value to match
    return_first: bool
        if True return the first objects found in the list else all the objects matching

    Returns
    -------
    list of tuple(object, int): object and index or list of object and indexes
    """
    selection = []
    obj = None
    for ind, obj_tmp in enumerate(objects):
        if find_object_if_matched_attr_name_val(obj_tmp, attr_name, attr_value):
            obj = obj_tmp
            if return_first:
                break
            else:
                selection.append((obj_tmp, ind))

    if obj is None:
        if return_first:
            return None, -1
        else:
            return []
    else:
        if return_first:
            return obj, ind
        else:
            return selection


def find_dict_if_matched_key_val(dict_tmp, key, value):
    """
    check if a key/value pair match in a given dictionnary
    Parameters
    ----------
    dict_tmp: (dict) the dictionnary to be tested
    key: (str) a key string to look for in dict_tmp
    value: (object) any python object

    Returns
    -------
    bool: True if the key/value pair has been found in dict_tmp

    """
    if key in dict_tmp:
        if dict_tmp[key] == value:
            return True
    return False


def find_dict_in_list_from_key_val(dicts, key, value, return_index=False):
    """ lookup within a list of dicts. Look for the dict within the list which has the correct key, value pair

    Parameters
    ----------
    dicts: (list) list of dictionnaries
    key: (str) specific key to look for in each dict
    value: value to match

    Returns
    -------
    dict: if found otherwise returns None
    """
    for ind, dict_tmp in enumerate(dicts):
        if find_dict_if_matched_key_val(dict_tmp, key, value):
            if return_index:
                return dict_tmp, ind
            else:
                return dict_tmp
    if return_index:
        return None, -1
    else:
        return None


def get_entrypoints(group='pymodaq.plugins'):
    """ Get the list of modules defined from a group entry point

    Because of evolution in the package, one or another of the forms below may be deprecated.
    We start from the newer way down to the older

    Parameters
    ----------
    group: str
        the name of the group
    """
    try:
        discovered_entrypoints = metadata.entry_points(group=group)
    except TypeError:
        try:
            discovered_entrypoints = metadata.entry_points().select(group=group)
        except AttributeError:
            discovered_entrypoints = metadata.entry_points().get(group, [])
    if isinstance(discovered_entrypoints, tuple):  # API for python > 3.8
        discovered_entrypoints = list(discovered_entrypoints)
    return discovered_entrypoints


@cache
def get_instrument_plugins():  # pragma: no cover
    """
    Get plugins names as a list
    Parameters
    ----------
    plugin_type: (str) plugin type either 'daq_0Dviewer', 'daq_1Dviewer', 'daq_2Dviewer', 'daq_NDviewer' or 'daq_move'
    module: (module) parent module of the plugins

    Returns
    -------

    """
    plugins_import = []
    discovered_plugins = []
    discovered_plugins_all = get_entrypoints(group='pymodaq.plugins')  # old naming of the instrument plugins
    discovered_plugins_all.extend(get_entrypoints(group='pymodaq.instruments'))  # new naming convention
    for entry in discovered_plugins_all:
        if entry.name not in [ent.name for ent in discovered_plugins]:
            discovered_plugins.append(entry)
    logger.debug(f'Found {len(discovered_plugins)} installed plugins, trying to import them')
    viewer_types = ['0D', '1D', '2D', 'ND']
    for entrypoint in discovered_plugins:
        #print(f'Looking for valid instrument plugins in package: {module.value}')
        plugin_list = []
        try:
            try:
                movemodule = importlib.import_module(f'{entrypoint.value}.daq_move_plugins', entrypoint.value)
                plugin_list.extend([{'name': mod[len('daq_move') + 1:],
                                     'module': movemodule,
                                     'parent_module': importlib.import_module(entrypoint.value),
                                     'type': 'daq_move'}
                                    for mod in [mod[1] for mod in pkgutil.iter_modules([str(movemodule.path.parent)])]
                                    if 'daq_move' in mod])
            except ModuleNotFoundError:
                pass
            viewer_modules = {}
            for vtype in viewer_types:
                try:
                    viewer_modules[vtype] = importlib.import_module(f'{entrypoint.value}.daq_viewer_plugins.plugins_{vtype}',
                                                    entrypoint.value)
                    plugin_list.extend([{'name': mod[len(f'daq_{vtype}viewer') + 1:],
                                     'module': viewer_modules[vtype],
                                     'parent_module': importlib.import_module(entrypoint.value),
                                     'type': f'daq_{vtype}viewer'}
                                    for mod in [mod[1] for mod in pkgutil.iter_modules([str(viewer_modules[vtype].path.parent)])]
                                    if f'daq_{vtype}viewer' in mod])
                except ModuleNotFoundError:
                    pass

            # check if modules are importable
            for mod in plugin_list:
                try:
                    plugin_type = mod['type']
                    if plugin_type == 'daq_move':
                        submodule = mod['module']
                        importlib.import_module(f'{submodule.__package__}.daq_move_{mod["name"]}')
                    else:
                        submodule = mod['module']
                        importlib.import_module(f'{submodule.__package__}.daq_{plugin_type[4:6]}viewer_{mod["name"]}')
                    plugins_import.append(mod)
                except Exception as e:  # pragma: no cover
                    """If an error is generated at the import, then exclude this plugin"""
                    logger.debug(f'Impossible to import Instrument plugin {mod["name"]}'
                                 f' from module: {mod["parent_module"].__package__}')
        except Exception as e:  # pragma: no cover
            logger.debug(str(e))
            
    # add utility plugin for PID
    try:
        pidmodule = importlib.import_module('pymodaq.extensions.pid')
        mod = [{'name': 'PID',
               'module': pidmodule,
               'parent_module': pidmodule,
               'type': 'daq_move'}]
        importlib.import_module(f'{pidmodule.__package__}.daq_move_PID')
        plugins_import.extend(mod)

    except Exception as e:
        logger.debug(f'Impossible to import PID utility plugin: {str(e)}')

    return plugins_import

def get_plugins(plugin_type='daq_0Dviewer'):  # pragma: no cover
    """
    Get plugins names as a list
    Parameters
    ----------
    plugin_type: (str) plugin type either 'daq_0Dviewer', 'daq_1Dviewer', 'daq_2Dviewer', 'daq_NDviewer' or 'daq_move'
    module: (module) parent module of the plugins

    Returns
    -------

    """
    return [plug for plug in get_instrument_plugins() if plug['type'] == plugin_type]


def check_vals_in_iterable(iterable1, iterable2):
    assert len(iterable1) == len(iterable2)
    iterable1 = list(iterable1)  # so the assertion below is valid for any kind of iterable, list, tuple, ndarray...
    iterable2 = list(iterable2)
    for val1, val2 in zip(iterable1, iterable2):
        assert val1 == val2


def caller_name(skip=2):
    """Get a name of a caller in the format module.class.method

       `skip` specifies how many levels of stack to skip while getting caller
       name. skip=1 means "who calls me", skip=2 "who calls my caller" etc.

       An empty string is returned if skipped levels exceed stack height
    """
    stack = inspect.stack()
    start = 0 + skip
    if len(stack) < start + 1:
        return ''
    parentframe = stack[start][0]

    name = []
    module = inspect.getmodule(parentframe)
    # `modname` can be None when frame is executed directly in console
    # TODO(techtonik): consider using __main__
    if module:
        name.append(module.__name__)
    # detect classname
    if 'self' in parentframe.f_locals:
        # I don't know any way to detect call from the object method
        # XXX: there seems to be no way to detect static method call - it will
        #      be just a function call
        name.append(parentframe.f_locals['self'].__class__.__name__)
    codename = parentframe.f_code.co_name
    if codename != '<module>':  # top level usually
        name.append(codename)  # function or a method
    del parentframe
    return ".".join(name)


def zeros_aligned(n, align, dtype=np.uint32):
    """
    Get aligned memory array wih alignment align.
    Parameters
    ----------
    n: (int) length in dtype bytes of memory
    align: (int) memory alignment
    dtype: (numpy.dtype) type of the stored memory elements

    Returns
    -------

    """
    dtype = np.dtype(dtype)
    nbytes = n * dtype.itemsize
    buff = np.zeros(nbytes + align, dtype=np.uint8)
    start_index = -buff.ctypes.data % align
    return buff[start_index:start_index + nbytes].view(dtype)


# ########################
# #File management

def get_new_file_name(base_path=Path(config('data_saving', 'h5file', 'save_path')), base_name='tttr_data'):
    if isinstance(base_path, str):
        base_path = Path(base_path)

    today = datetime.datetime.now()

    date = today.strftime('%Y%m%d')
    year = today.strftime('%Y')
    year_dir = base_path.joinpath(year)
    if not year_dir.is_dir():
        year_dir.mkdir()
    curr_dir = base_path.joinpath(year, date)
    if not curr_dir.is_dir():
        curr_dir.mkdir()

    files = []
    for entry in curr_dir.iterdir():
        if entry.name.startswith(base_name) and entry.is_file():
            files.append(entry.stem)
    files.sort()
    if not files:
        index = 0
    else:
        index = int(files[-1][-3:]) + 1

    file = f'{base_name}_{index:03d}'
    return file, curr_dir


# ##############
# Math utilities
# math utility functions, should now be imported from the math_utils module
import pymodaq.utils.math_utils as mutils

def my_moment(x, y):
    deprecation_msg(f'my_moment function should now be imported from the {mutils.__name__} module', stacklevel=3)
    return mutils.my_moment(x, y)

def normalize(x):
    deprecation_msg(f'normalize function should now be imported from the {mutils.__name__} module', stacklevel=3)
    return mutils.normalize(x)


def odd_even(x):
    deprecation_msg(f'odd_even function should now be imported from the {mutils.__name__} module', stacklevel=3)
    return mutils.odd_even(x)


def greater2n(x):
    deprecation_msg(f'greater2n function should now be imported from the {mutils.__name__} module', stacklevel=3)
    return mutils.greater2n(x)


def linspace_step(start, stop, step):
    deprecation_msg(f'linspace_step function should now be imported from the {mutils.__name__} module', stacklevel=3)
    return mutils.linspace_step(start, stop, step)


def linspace_step_N(start, step, Npts):
    deprecation_msg(f'linspace_step_N function should now be imported from the {mutils.__name__} module', stacklevel=3)
    return mutils.linspace_step_N(start, step, Npts)


def find_index(x, threshold):
    deprecation_msg(f'find_index function should now be imported from the {mutils.__name__} module', stacklevel=3)
    return mutils.find_index(x, threshold)


def find_common_index(x, y, x0, y0):
    deprecation_msg(f'find_common_index function should now be imported from the {mutils.__name__} module', stacklevel=3)
    return mutils.find_common_index(x, y, x0, y0)


def gauss1D(x, x0, dx, n=1):
    deprecation_msg(f'gauss1D function should now be imported from the {mutils.__name__} module', stacklevel=3)
    return mutils.gauss1D(x, x0, dx, n=n)


def gauss2D(x, x0, dx, y, y0, dy, n=1, angle=0):
    deprecation_msg(f'gauss2D function should now be imported from the {mutils.__name__} module', stacklevel=3)
    return mutils.gauss2D(x, x0, dx, y, y0, dy, n, angle)

def ftAxis(Npts, omega_max):
    deprecation_msg(f'ftAxis function should now be imported from the {mutils.__name__} module', stacklevel=3)
    return mutils.ftAxis(Npts, omega_max)


def ftAxis_time(Npts, time_max):
    deprecation_msg(f'ftAxis_time function should now be imported from the {mutils.__name__} module', stacklevel=3)
    return mutils.ftAxis_time(Npts, time_max)


def ft(x, dim=-1):
    deprecation_msg(f'ft function should now be imported from the {mutils.__name__} module', stacklevel=3)
    return mutils.ft(x, dim)


def ift(x, dim=0):
    deprecation_msg(f'ift function should now be imported from the {mutils.__name__} module', stacklevel=3)
    return mutils.ift(x, dim)


def ft2(x, dim=(-2, -1)):
    deprecation_msg(f'ft2 function should now be imported from the {mutils.__name__} module', stacklevel=3)
    return mutils.ft2(x, dim)


def ift2(x, dim=(-2, -1)):
    deprecation_msg(f'ift2 function should now be imported from the {mutils.__name__} module', stacklevel=3)
    return mutils.ift2(x, dim)




if __name__ == '__main__':
    #paths = recursive_find_expr_in_files('C:\\Users\\weber\\Labo\\Programmes Python\\PyMoDAQ_Git', 'visa')
    # for p in paths:
    #     print(str(p))
    # v = get_version()
    # pass
    #plugins = get_plugins()  # pragma: no cover
    #extensions = get_extension()
    #models = get_models()
    #count = count_lines('C:\\Users\\weber\\Labo\\Programmes Python\\PyMoDAQ_Git\\pymodaq\src')


    # import license
    # mit = license.find('MIT')
    #

    # paths = recursive_find_expr_in_files(r'C:\Users\weber\Labo\Programmes Python\PyMoDAQ_Git',
    #                                      exp="cfunc",
    #                                      paths=[],
    #                                      filters=['.git', '.idea', '__pycache__', 'build', 'egg', 'documentation',
    #                                               '.tox',],
    #                                      replace=False,
    #                                      replace_str="pymodaq.utils")

    #get_version()
    get_instrument_plugins()
    #get_plugins('daq_move')
    #get_plugins('daq_0Dviewer')
    pass

    # paths = recursive_find_files('C:\\Users\\weber\\Labo\\Programmes Python\\PyMoDAQ_Git',
    #                      exp='VERSION', paths=[])
    # import version
    # for file in paths:
    #     with open(str(file), 'r') as f:
    #         v = version.Version(f.read())
    #         v.minor += 1
    #         v.patch = 0
    #     with open(str(file), 'w') as f:
    #         f.write(str(v))

    # for file in paths:
    #     with open(str(file), 'w') as f:
    #         f.write(mit.render(name='Sebastien Weber', email='sebastien.weber@cemes.fr'))




from typing import List, Union, TYPE_CHECKING, Iterable
import numpy as np

from qtpy import QtWidgets
from qtpy.QtCore import QObject, Signal, Slot

from pyqtgraph.graphicsItems import InfiniteLine, ROI

from pymodaq.utils.logger import set_logger, get_module_name
from pymodaq.utils.data import DataToExport, DataRaw, DataWithAxes, Axis, DataSource
from pymodaq.utils.exceptions import ViewerError
from pymodaq.utils.enums import BaseEnum, enum_checker
from pymodaq.utils.factory import ObjectFactory, BuilderBase
from pymodaq.utils.plotting import data_viewers
from pymodaq.utils.gui_utils import DockArea, Dock
from pymodaq.utils.managers.parameter_manager import ParameterManager

if TYPE_CHECKING:
    from pymodaq.utils.plotting.data_viewers.viewer0D import Viewer0D
    from pymodaq.utils.plotting.data_viewers.viewer1D import Viewer1D
    from pymodaq.utils.plotting.data_viewers.viewer2D import Viewer2D
    from pymodaq.utils.plotting.data_viewers.viewerND import ViewerND

config_viewers = {
}

logger = set_logger(get_module_name(__file__))


class ViewersEnum(BaseEnum):
    """enum relating a given viewer with data type"""
    Viewer0D = 'Data0D'
    Viewer1D = 'Data1D'
    Viewer2D = 'Data2D'
    ViewerND = 'DataND'
    ViewerSequential = 'DataSequential'

    def get_dim(self):
        return self.value.split('Data')[1].split('D')[0]

    def increase_dim(self, ndim: int):
        dim = self.get_dim()
        if dim != 'N':
            dim_as_int = int(dim) + ndim
            if dim_as_int > 2:
                dim = 'N'
            else:
                dim = str(dim_as_int)
        else:
            dim = 'N'
        return ViewersEnum[f'Viewer{dim}D']


def get_viewer_enum_from_axes(Naxes: int):
    if Naxes < 0:
        raise ValueError('Naxes could not be below 0')
    if Naxes == 0:
        viewer_enum = ViewersEnum['Viewer0D']
    elif Naxes == 1:
        viewer_enum = ViewersEnum['Viewer1D']
    elif Naxes == 2:
        viewer_enum = ViewersEnum['Viewer2D']
    else:
        viewer_enum = ViewersEnum['ViewerND']
    return viewer_enum


class ViewerFactory(ObjectFactory):
    def get(self, viewer_name, **kwargs):
        if viewer_name not in ViewersEnum.names():
            raise ValueError(f'{viewer_name} is not a valid PyMoDAQ Viewer: {ViewersEnum.names()}')
        return self.create(viewer_name, **kwargs)

    @property
    def viewers(self):
        return self.keys


@ViewerFactory.register('Viewer0D')
def create_viewer0D(parent: QtWidgets.QWidget = None, **_ignored):
    return data_viewers.viewer0D.Viewer0D(parent)


@ViewerFactory.register('Viewer1D')
def create_viewer1D(parent: QtWidgets.QWidget, **_ignored):
    return data_viewers.viewer1D.Viewer1D(parent)


@ViewerFactory.register('Viewer2D')
def create_viewer2D(parent: QtWidgets.QWidget, **_ignored):
    return data_viewers.viewer2D.Viewer2D(parent)


@ViewerFactory.register('ViewerND')
def create_viewerND(parent: QtWidgets.QWidget, **_ignored):
    return data_viewers.viewerND.ViewerND(parent)


# @ViewerFactory.register('ViewerSequential')
# def create_viewer_sequential(widget: QtWidgets.QWidget, **_ignored):
#     return data_viewers.viewer_sequential.ViewerSequential(widget)


viewer_factory = ViewerFactory()


class ViewerBase(QObject):
    """Base Class for data viewers implementing all common functionalities

    Parameters
    ----------
    parent: QtWidgets.QWidget
    title: str

    Attributes
    ----------
    view: QObject
        Ui interface of the viewer

    data_to_export_signal: Signal[DataToExport]
    ROI_changed: Signal
    crosshair_dragged: Signal[float, float]
    crosshair_clicked: Signal[bool]
    sig_double_clicked: Signal[float, float]
    status_signal: Signal[str]
    """
    data_to_export_signal = Signal(DataToExport)
    _data_to_show_signal = Signal(DataWithAxes)

    ROI_changed = Signal()
    crosshair_dragged = Signal(float, float)  # Crosshair position in units of scaled top/right axes
    status_signal = Signal(str)
    crosshair_clicked = Signal(bool)
    sig_double_clicked = Signal(float, float)

    def __init__(self, parent: QtWidgets.QWidget = None, title=''):
        super().__init__()
        self.title = title if title != '' else self.__class__.__name__

        self._raw_data = None
        self.data_to_export: DataToExport = DataToExport(name=self.title)
        self.view: Union[Viewer0D, Viewer1D, Viewer2D, ViewerND] = None

        if parent is None:
            parent = QtWidgets.QWidget()
            parent.show()
        self.parent = parent

        self._display_temporary = False

    @property
    def has_action(self):
        """Convenience method"""
        if hasattr(self.view, 'has_action'):
            return self.view.has_action

    @property
    def is_action_checked(self):
        """Convenience method"""
        if hasattr(self.view, 'is_action_checked'):
            return self.view.is_action_checked

    @property
    def is_action_visible(self):
        """Convenience method"""
        if hasattr(self.view, 'is_action_visible'):
            return self.view.is_action_visible

    @property
    def set_action_checked(self):
        """Convenience method"""
        if hasattr(self.view, 'set_action_checked'):
            return self.view.set_action_checked

    @property
    def set_action_visible(self):
        """Convenience method"""
        if hasattr(self.view, 'set_action_visible'):
            return self.view.set_action_visible

    @property
    def get_action(self):
        """Convenience method"""
        if hasattr(self.view, 'get_action'):
            return self.view.get_action

    @property
    def toolbar(self):
        """Convenience property"""
        if hasattr(self.view, 'toolbar'):
            return self.view.toolbar

    @property
    def viewer_type(self):
        """str: the viewer data type see DATA_TYPES"""
        return ViewersEnum[self.__class__.__name__].value

    def show_data(self, data: DataWithAxes, **kwargs):
        """Entrypoint to display data into the viewer

        Parameters
        ----------
        data: data_mod.DataFromPlugins
        """
        if len(data.shape) > 4:
            raise ViewerError(f'Ndarray of dim: {len(data.shape)} cannot be plotted using a {self.viewer_type}')

        self.data_to_export = DataToExport(name=self.title)
        self._raw_data = data

        self._display_temporary = False

        self._show_data(data, **kwargs)

    def show_data_temp(self, data: DataRaw, **kwargs):
        """Entrypoint to display temporary data into the viewer

        No processed data signal is emitted from the viewer

        Parameters
        ----------
        data: data_mod.DataFromPlugins
        """
        self._display_temporary = True
        self.show_data(data, **kwargs)

    def _show_data(self, data: DataRaw):
        """Specific viewers should implement it"""
        raise NotImplementedError

    def add_attributes_from_view(self):
        """Convenience function to add attributes from the view to self"""
        for attribute in self.convenience_attributes:
            if hasattr(self.view, attribute):
                setattr(self, attribute, getattr(self.view, attribute))

    def trigger_action(self, action_name: str):
        """Convenience function to trigger programmatically one of the action of the related view"""
        if self.has_action(action_name):
            self.get_action(action_name).trigger()

    def activate_roi(self, activate=True):
        """Activate the Roi manager using the corresponding action"""
        raise NotImplementedError

    def setVisible(self, show=True):
        """convenience method to show or hide the paretn widget"""
        self.parent.setVisible(show)

    @property
    def roi_target(self) -> Union[InfiniteLine.InfiniteLine, ROI.ROI]:
        """To be implemented if necessary (Viewer1D and above)"""
        return None

    def move_roi_target(self, pos: Iterable[float] = None, **kwargs):
        """move a specific read only ROI at the given position on the viewer"""
        ...

    def show_roi_target(self, show=True):
        """Show/Hide a specific read only ROI"""
        if self.roi_target is not None:
            self.roi_target.setVisible(show)


class ViewerDispatcher:
    """MixIn class to add easy control for adding multuiple data viewers in docks depending on data to be plotted"""

    def __init__(self, dockarea: DockArea = None, title: str = '', next_to_dock: Dock = None):
        super().__init__()
        self._title = title
        self._next_to_dock = next_to_dock
        if dockarea is None:
            dockarea = DockArea()
            dockarea.show()
        self.dockarea = dockarea

        self._viewer_docks = []
        self._viewer_widgets = []
        self._viewer_types = []
        self._viewers = []

    @property
    def viewers(self) -> List[ViewerBase]:
        return self._viewers

    @property
    def viewer_docks(self) -> List[Dock]:
        return self._viewer_docks

    @property
    def viewer_widgets(self) -> List[QtWidgets.QWidget]:
        return self._viewer_widgets

    @property
    def viewer_types(self) -> List[ViewersEnum]:
        return self._viewer_types

    def remove_viewers(self, Nviewers_to_leave: int = 0):
        """Remove viewers from the list after index Nviewers_to_leave

        Parameters
        ----------
        Nviewers

        Returns
        -------

        """
        while len(self.viewer_docks) > Nviewers_to_leave:
            widget = self.viewer_widgets.pop()
            widget.close()
            dock = self.viewer_docks.pop()
            dock.close()
            self.viewers.pop()
            self.viewer_types.pop()
            QtWidgets.QApplication.processEvents()

    def add_viewer(self, viewer_type: ViewersEnum, dock_viewer=None, dock_name=None):
        viewer_type = enum_checker(ViewersEnum, viewer_type)

        if dock_viewer is None:
            if dock_name is None:
                dock_name = f'{self._title}_Viewer_{len(self.viewer_docks) + 1}'
            dock_viewer = Dock(dock_name, size=(350, 350), closable=False)
        self.viewer_docks.append(dock_viewer)

        self._viewer_widgets.append(QtWidgets.QWidget())
        self.viewers.append(viewer_factory.get(viewer_type.name, parent=self._viewer_widgets[-1]))

        self.viewer_types.append(viewer_type)

        self.viewer_docks[-1].addWidget(self._viewer_widgets[-1])
        # if len(self.viewer_docks) == 1:
        #     if self._next_to_dock is not None:
        #         self.dockarea.addDock(self.viewer_docks[-1], 'right', self._next_to_dock)
        #     else:
        #         self.dockarea.addDock(self.viewer_docks[-1])
        # else:
        #     self.dockarea.addDock(self.viewer_docks[-1], 'right', self.viewer_docks[-2])
        self.dockarea.addDock(self.viewer_docks[-1], 'right')

    def update_viewers(self, viewers_type: List[ViewersEnum], viewers_name: List[str] = None, force=False):
        """

        Parameters
        ----------
        viewers_type: List[ViewersEnum]
        viewers_name: List[str] or None
        force: bool
            if True remove all viewers before update else check if new viewers type are compatible with old ones

        Returns
        -------

        """

        Nviewers_to_leave = 0
        if not force:
            # check if viewers are compatible with new data dim
            for ind, viewer_type in enumerate(viewers_type):
                if len(self.viewer_types) > ind:
                    if viewer_type == self.viewer_types[ind]:
                        Nviewers_to_leave += 1
                    else:
                        break
                else:
                    break
        self.remove_viewers(Nviewers_to_leave)
        ind_loop = 0
        while len(self.viewers) < len(viewers_type):
            self.add_viewer(viewers_type[Nviewers_to_leave + ind_loop],
                            dock_name=viewers_name[Nviewers_to_leave + ind_loop] if viewers_name is not None else None)
            ind_loop += 1
        QtWidgets.QApplication.processEvents()

    def close(self):
        for dock in self.viewer_docks:
            dock.close()

    def show_data(self, data: DataToExport):
        """ Convenience method. Display each dwa in a dedicated data viewer"""
        viewer_types = [ViewersEnum(dwa.dim.name) for dwa in data]
        if self.viewer_types != viewer_types:
            self.update_viewers(viewer_types)
        for viewer, dwa in zip(self.viewers, data):
            if len(dwa.axes) != len(dwa.shape):
                dwa.create_missing_axes()
            viewer.show_data(dwa)


if __name__ == '__main__':

    LABEL = 'A Label'
    UNITS = 'units'
    OFFSET = -20.4
    SCALING = 0.22
    SIZE = 20
    DATA = OFFSET + SCALING * np.linspace(0, SIZE - 1, SIZE)

    DATA0D = np.array([2.7])
    DATA1D = np.arange(0, 10)
    DATA2D = np.arange(0, 5 * 6).reshape((5, 6))
    DATAND = np.arange(0, 5 * 6 * 3).reshape((5, 6, 3))
    Nn0 = 10
    Nn1 = 5


    def init_axis(data=None, index=0):
        if data is None:
            data = DATA
        return Axis(label=LABEL, units=UNITS, data=data, index=index)


    def init_data(data=None, Ndata=1, axes=[], name='myData', source=DataSource['raw'],
                  labels=None) -> DataWithAxes:
        if data is None:
            data = DATA2D
        return DataWithAxes(name, source, data=[data for ind in range(Ndata)],
                            axes=axes, labels=labels)


    def ini_data_to_export():
        dat1 = init_data(data=DATA2D, Ndata=2, name='data2D')
        dat2 = init_data(data=DATA1D, Ndata=3, name='data1D')
        data = DataToExport(name='toexport', data=[dat1, dat2])
        return dat1, dat2, data

    import sys
    import random
    app = QtWidgets.QApplication(sys.argv)

    dockarea = DockArea()
    prog = ViewerDispatcher(dockarea=dockarea, title='Dispatcher')
    dockarea.show()

    _, _, dte = ini_data_to_export()

    # N = 2
    # viewers = ['Viewer0D', 'Viewer1D', 'Viewer2D']
    # viewers = [ViewersEnum[random.choice(viewers)] for ind in range(N)]
    # viewers = ['Viewer2D', 'Viewer2D', ]
    # print(viewers)
    # prog.update_viewers(viewers)

    prog.show_data(dte)

    sys.exit(app.exec_())



# -*- coding: utf-8 -*-
"""
Created the 05/12/2022

@author: Sebastien Weber
"""
from typing import List, Tuple, Any
import re
import numpy as np
from pymodaq.utils.data import Axis, DataDistribution
from pymodaq.utils.logger import set_logger, get_module_name
from pymodaq.utils import math_utils as mutils
from pymodaq.utils import config as configmod
from pymodaq.utils.plotting.scan_selector import Selector

from ..scan_factory import ScannerFactory, ScannerBase, ScanParameterManager


logger = set_logger(get_module_name(__file__))
config = configmod.Config()


class Scan1DBase(ScannerBase):
    scan_type = 'Scan1D'

    params = []
    n_axes = 1
    distribution = DataDistribution['uniform']

    def __init__(self, actuators: List = None, **_ignored):
        super().__init__(actuators=actuators)

    def get_nav_axes(self) -> List[Axis]:
        return [Axis(label=f'{self.actuators[0].title}',
                     units=f'{self.actuators[0].units}',
                     data=np.squeeze(self.positions))]

    def get_scan_shape(self) -> Tuple[int]:
        return len(self.positions),

    def get_indexes_from_scan_index(self, scan_index: int) -> Tuple[int]:
        """To be reimplemented. Calculations of indexes within the scan"""
        return (scan_index,)

    def update_from_scan_selector(self, scan_selector: Selector):
        pass


@ScannerFactory.register()
class Scan1DLinear(Scan1DBase):
    """ Defines a linear scan between start and stop values with steps of length defined in the step setting"""

    scan_subtype = 'Linear'
    params = [
        {'title': 'Start:', 'name': 'start', 'type': 'float', 'value': 0.},
        {'title': 'Stop:', 'name': 'stop', 'type': 'float', 'value': 1.},
        {'title': 'Step:', 'name': 'step', 'type': 'float', 'value': 0.1}
        ]
    n_axes = 1
    distribution = DataDistribution['uniform']

    def __init__(self, actuators: List = None, **_ignored):
        ScannerBase.__init__(self, actuators=actuators)

    def set_scan(self):
        self.positions = mutils.linspace_step(self.settings['start'], self.settings['stop'],
                                              self.settings['step'])
        self.get_info_from_positions(self.positions)

    def set_settings_titles(self):
        if len(self.actuators) == 1:
            self.settings.child('start').setOpts(title=f'{self.actuators[0].title} start:')
            self.settings.child('stop').setOpts(title=f'{self.actuators[0].title} stop:')
            self.settings.child('step').setOpts(title=f'{self.actuators[0].title} step:')

    def evaluate_steps(self) -> int:
        n_steps = int(np.abs((self.settings['stop'] - self.settings['start']) / self.settings['step']) + 1)
        return n_steps

    def update_from_scan_selector(self, scan_selector: Selector):
        coordinates = scan_selector.get_coordinates()
        if coordinates.shape == (2, 2) or coordinates.shape == (2, 1):
            self.settings.child('start').setValue(coordinates[0, 0])
            self.settings.child('stop').setValue(coordinates[1, 0])


@ScannerFactory.register()
class Scan1DRandom(Scan1DLinear):
    """ Defines a  random linear scan by first initializing a linear one between start and stop values with
    steps of length defined in the step setting, then shuffling the values."""

    scan_subtype = 'Random'

    def __init__(self, actuators: List = None, **_ignored):
        super().__init__(actuators=actuators)

    def set_scan(self):
        self.positions = mutils.linspace_step(self.settings['start'], self.settings['stop'],
                                              self.settings['step'])
        np.random.shuffle(self.positions)
        self.get_info_from_positions(self.positions)
        self.set_settings_titles()

        
@ScannerFactory.register()
class Scan1DSparse(Scan1DBase):
    """ Syntax goes as start:step:stop or with single entry

    * 0:0.2:1 will give [0 0.2 0.4 0.6 0.8 1]
    * 0 will give [0]

    Separate entries with comma or new line:

    * 0:0.2:1,5 will give [0 0.2 0.4 0.6 0.8 1 5]
    * 0:0.2:1,5:1:7 will give [0 0.2 0.4 0.6 0.8 1 5 6 7]
    """

    scan_subtype = 'Sparse'
    params = [
        {'title': 'Parsed string:', 'name': 'parsed_string', 'type': 'text', 'value': '0:0.1:1', }
        ]
    n_axes = 1
    distribution = DataDistribution['uniform']  # because in 1D it doesn't matter is spread or
    # uniform, one can easily plot both types on a regulat 1D plot

    def __init__(self, actuators: List = None, **_ignored):
        super().__init__(actuators=actuators)
        self.settings.child('parsed_string').setOpts(tip=self.__doc__)

    def set_scan(self):
        try:
            range_strings = re.findall("[^,\s]+", self.settings['parsed_string'])
            series = np.asarray([])
            for range_string in range_strings:
                number_strings = re.findall("[^:]+", range_string)  # Extract the numbers by splitting on :.
                this_range = np.asarray([])
                if len(number_strings) == 3:  # 3 Numbers specify a range
                    start, step, stop = [float(number) for number in number_strings]
                    this_range = mutils.linspace_step(start, stop, step)
                elif len(number_strings) == 1:  # 1 number just specifies a single number
                    this_range = np.asarray([float(number_strings[0])])
                series = np.concatenate((series, this_range))

            self.positions = np.atleast_1d(np.squeeze(series))
            self.get_info_from_positions(self.positions)
        except Exception as e:
            pass  # many things could happen when parsing strings

    def set_settings_titles(self):
        if len(self.actuators) == 1:
            self.settings.child('start').setOpts(title=f'{self.actuators[0].title} start:')

    def evaluate_steps(self) -> int:
        """Quick evaluation of the number of steps to stop the calculation if the evaluation os above the
        configured limit"""
        self.set_scan()  # no possible quick evaluation, easiest to process it
        return self.n_steps

    def set_settings_titles(self):
        if len(self.actuators) == 1:
            self.settings.child('parsed_string').setOpts(title=f'{self.actuators[0].title} Parsed string:')

try:
    import adaptive


    @ScannerFactory.register()
    class Scan1DAdaptive(Scan1DBase):

        scan_subtype = 'Adaptive'
        params = [
            {'title': 'Loss type', 'name': 'scan_loss', 'type': 'list',
             'limits': ['default', 'curvature', 'uniform'], 'tip': 'Type of loss used by the algo. to determine next points'},
            {'title': 'Start:', 'name': 'start', 'type': 'float', 'value': 0.},
            {'title': 'Stop:', 'name': 'stop', 'type': 'float', 'value': 1.},
            ]
        distribution = DataDistribution['spread']

        def __init__(self, actuators: List = None, **_ignored):
            super().__init__(actuators=actuators)

        def set_scan(self):
            self.axes_unique = [np.array([])]
            self.axes_indexes = np.array([], dtype=int)
            self.positions = np.array([self.settings['start'], self.settings['stop']])

        def evaluate_steps(self) -> int:
            return 1

        def get_nav_axes(self) -> List[Axis]:
            return [Axis(label=f'{self.actuators[0].mod_name} axis',
                         units=f'{self.actuators[0].units}',
                         data=self.positions[0])]

        def get_scan_shape(self) -> Tuple[int]:
            return len(self.positions),

except ModuleNotFoundError:
    logger.info('adaptive module is not present, no adaptive scan possible')




# -*- coding: utf-8 -*-
"""
Created the 02/03/2023

@author: Sebastien Weber
"""

import numpy as np


# project imports
from pymodaq.utils.h5modules.backends import Node
from pymodaq.utils.h5modules.exporter import ExporterFactory, H5Exporter
from pymodaq.utils.h5modules.backends import H5Backend


@ExporterFactory.register_exporter()
class H5h5Exporter(H5Exporter):
    """ Exporter object for saving nodes as single h5 files"""

    FORMAT_DESCRIPTION = "Single node h5 file"
    FORMAT_EXTENSION = "h5"

    def export_data(self, node: Node, filename: str) -> None:
        """Export an h5 node"""
        # This should allow to get the base file object
        if node.backend == 'tables':
            basefile = node.node._v_file
            basefile.copy_file(dstfilename=str(filename), overwrite=False)
        else:
            import h5py
            with h5py.File(filename, 'w') as f_dest:
                node.node.h5file.copy(self.h5file, f_dest)

        # basefile = node.get_file()
        # basefile.copy_file(dstfilename=str(filename), overwrite=False)

        new_file = H5Backend(backend="tables")
        new_file.open_file(str(filename), 'a')

        new_file.h5file.move_node(self.get_node_path(node), newparent=new_file.h5file.get_node('/'))
        new_file.h5file.remove_node('/RawData', recursive=True)
        new_file.close_file()


@ExporterFactory.register_exporter()
class H5txtExporter(H5Exporter):
    """ Exporter object for saving nodes as txt files"""

    FORMAT_DESCRIPTION = "Text files"
    FORMAT_EXTENSION = "txt"

    def export_data(self, node: Node, filename: str) -> None:
        """Export the node as a .txt file format"""
        if 'ARRAY' in node.attrs['CLASS']:
            data = node.read()
            if not isinstance(data, np.ndarray):
                # in case one has a list of same objects (array of strings for instance, logger or other)
                data = np.array(data)
                np.savetxt(filename, data, '%s', '\t')
            else:
                np.savetxt(filename, data, '%.6e', '\t')
        elif 'GROUP' in node.attrs['CLASS']:
            data_tot = []
            header = []
            dtypes = []
            fmts = []
            for subnode_name, subnode in node.children().items():
                if 'ARRAY' in subnode.attrs['CLASS']:
                    if len(subnode.attrs['shape']) == 1:
                        data = subnode.read()
                        if not isinstance(data, np.ndarray):
                            # in case one has a list of same objects (array of strings for instance, logger or other)
                            data = np.array(data)
                        data_tot.append(data)
                        dtypes.append((subnode_name, data.dtype))
                        header.append(subnode_name)
                        if data.dtype.char == 'U':
                            fmt = '%s'  # for strings
                        elif data.dtype.char == 'l':
                            fmt = '%d'  # for integers
                        else:
                            fmt = '%.6f'  # for decimal numbers
                        fmts.append(fmt)

            data_trans = np.array(list(zip(*data_tot)), dtype=dtypes)
            np.savetxt(filename, data_trans, fmts, '\t', header='#' + '\t'.join(header))


@ExporterFactory.register_exporter()
class H5npyExporter(H5Exporter):
    """ Exporter object for saving nodes as npy files"""

    FORMAT_DESCRIPTION = "Binary NumPy format"
    FORMAT_EXTENSION = "npy"

    def export_data(self, node: Node, filename: str) -> None:
        """Export the node as a numpy binary file format"""
        # String __contain__ method will evaluate to True for CARRAY,EARRAY,VLARRAY,stringARRAY
        if 'ARRAY' in node.attrs['CLASS']:
            data = node.read()
            if not isinstance(data, np.ndarray):
                data = np.array(data)

            np.save(filename, data)


# -*- coding: utf-8 -*-
"""
Created the 06/01/2023

@author: Sebastien Weber
"""

import sys
from qtpy import QtWidgets, QtCore, QtSvg
from pyqtgraph.widgets.GraphicsView import GraphicsView


class SVGView:

    def __init__(self, parent: QtWidgets.QWidget = None):
        if parent is None:
            parent = QtWidgets.QWidget()
        self.parent_widget = parent
        self.parent_widget.setLayout(QtWidgets.QHBoxLayout())
        self.graphicsView = GraphicsView()
        self.parent_widget.layout().addWidget(self.graphicsView)

        svg_item = QtSvg.QGraphicsSvgItem(r'C:\Users\weber\Labo\Projet-Dossier candidature\Technical project\GDSII\wafer.svg')

        self.graphicsView.sceneObj.addItem(svg_item)


if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)

    widget = QtWidgets.QWidget()
    prog = SVGView(widget)
    widget.show()

    sys.exit(app.exec_())



# -*- coding: utf-8 -*-
"""
Created the 27/10/2022

@author: Sebastien Weber
"""
import importlib
from pathlib import Path
import pkgutil

from pymodaq.utils import logger as logger_module
from pymodaq.utils.daq_utils import get_entrypoints
logger = logger_module.set_logger(logger_module.get_module_name(__file__))


def get_ext_modules(path: Path):
    modules = []
    for mod in pkgutil.iter_modules([path]):
        modules.append(mod.name)
    return modules


def get_extensions():
    """
    Get pymodaq extensions as a list

    Returns
    -------
    list: list of dict containing the name and module of the found extension

        Each dict is defined with four keys:
            * pkg: the name of the plugin package
            * module: the module name where your extension class is defined
            * class_name: the name of the class defining the extension
            * name: a nice name for your extension to be displayed in the menu

    """
    extension_import = []
    discovered_extension = get_entrypoints(group='pymodaq.extensions')
    if len(discovered_extension) > 0:
        for pkg in discovered_extension:
            try:
                module = importlib.import_module(pkg.value)
                modules = get_ext_modules(Path(module.__path__[0]).joinpath('extensions'))
                for mod in modules:
                    mod_in = importlib.import_module(f'{pkg.value}.extensions.{mod}')
                    if hasattr(mod_in, 'EXTENSION_NAME'):
                        extension_import.append({'pkg': pkg.value, 'module': mod, 'name': mod_in.EXTENSION_NAME,
                                                 'class_name': mod_in.CLASS_NAME})

            except Exception as e:  # pragma: no cover
                logger.warning(f'Impossible to import the {pkg.value}.extensions.{mod} extension: {str(e)}')

    return extension_import



from qtpy import QtWidgets
from pyqtgraph.parametertree.parameterTypes.basetypes import WidgetParameterItem


class BoolPushParameterItem(WidgetParameterItem):
    """Registered parameter type which displays a QLineEdit"""

    def makeWidget(self):
        opts = self.param.opts
        w = QtWidgets.QPushButton()
        if 'label' in opts:
            w.setText(opts['label'])
        elif 'title' in opts:
            w.setText(opts['title'])
        else:
            w.setText(opts['name'])
        # w.setMaximumWidth(50)
        w.setCheckable(True)
        w.sigChanged = w.toggled
        w.value = w.isChecked
        w.setValue = w.setChecked
        w.setEnabled(not opts.get('readonly', False))
        self.hideWidget = False
        return w




from typing import List

import pymodaq.utils.config
import pymodaq.utils.gui_utils.dock
import pymodaq.utils.gui_utils.file_io
import pymodaq.utils.messenger
from qtpy import QtWidgets, QtCore
import sys
import os

from pymodaq.utils.managers.parameter_manager import ParameterManager
from pymodaq.utils.managers.modules_manager import ModulesManager
from pymodaq.utils.gui_utils import Dock
from pymodaq.utils.logger import set_logger, get_module_name
from pymodaq.utils import config as config_mod
from pymodaq.utils.parameter import ioxml
from pymodaq.utils.scanner import Scanner
from pymodaq.utils.scanner.scan_factory import ScannerBase
from pymodaq.utils.scanner.utils import ScanType #, adaptive_losses
from pathlib import Path
from collections import OrderedDict
from pymodaq.utils.messenger import messagebox

logger = set_logger(get_module_name(__file__))

batch_path = config_mod.get_set_batch_path()

params = [
    {'title': 'Actuators/Detectors Selection', 'name': 'modules', 'type': 'group', 'children': [
        {'title': 'detectors', 'name': 'detectors', 'type': 'itemselect'},
        {'title': 'Actuators', 'name': 'actuators', 'type': 'itemselect'},
    ]},
    ]


class BatchManager(ParameterManager):
    settings_name = 'batch_settings'
    params = [{'title': 'Filename:', 'name': 'filename', 'type': 'str', 'value': 'batch_default'},
              {'title': 'Scans', 'name': 'scans', 'type': 'group', 'children': []}]

    def __init__(self, msgbox=False, actuators=[], detectors=[], path=None):
        super().__init__()

        self.modules_manager: ModulesManager = ModulesManager(detectors, actuators)
        self.modules_manager.show_only_control_modules(True)
        self.modules_manager.actuators_changed[list].connect(self.update_actuators)
        self.modules_manager.settings_tree.setMinimumHeight(200)
        self.modules_manager.settings_tree.setMaximumHeight(200)

        self._scans = OrderedDict([])

        self.scanner = Scanner(actuators=self.modules_manager.actuators_all)

        self.settings_tree.setMinimumWidth(400)
        self.settings_tree.setMaximumWidth(500)
        self.settings_tree.setMinimumHeight(500)

        if path is None:
            path = batch_path
        else:
            assert isinstance(path, Path)
        self.batch_path = path

        if msgbox:
            msgBox = QtWidgets.QMessageBox()
            msgBox.setText("Scan Batch Manager?")
            msgBox.setInformativeText("What do you want to do?")
            cancel_button = msgBox.addButton(QtWidgets.QMessageBox.Cancel)
            new_button = msgBox.addButton("New", QtWidgets.QMessageBox.ActionRole)
            modify_button = msgBox.addButton('Modify', QtWidgets.QMessageBox.AcceptRole)
            msgBox.setDefaultButton(QtWidgets.QMessageBox.Cancel)
            ret = msgBox.exec()

            if msgBox.clickedButton() == new_button:
                self.set_new_batch()

            elif msgBox.clickedButton() == modify_button:
                self.set_file_batch()
            else:  # cancel
                pass

    def get_act_dets(self):
        acts = dict([])
        dets = dict([])
        for name in [child.name() for child in self.settings.child('scans').children()]:
            acts[name] = self.settings.child('scans', name, 'modules', 'actuators').value()['selected']
            dets[name] = self.settings.child('scans', name, 'modules', 'detectors').value()['selected']
        return acts, dets

    def set_file_batch(self, filename=None, show=True):
        """

        """
        if filename is None or filename is False:
            filename = pymodaq.utils.gui_utils.file_io.select_file(start_path=self.batch_path, save=False, ext='xml')
            if filename == '':
                return

        status = False
        settings_tmp = self.create_parameter(filename)
        children = settings_tmp.child('scans').children()

        #self.settings = self.create_parameter(self.params)
        actuators = children[0].child('modules', 'actuators').value()['all_items']
        if actuators != self.modules_manager.actuators_name:
            messagebox(text='The loaded actuators from the batch file do not corresponds to the dashboard actuators')
            return

        detectors = children[0].child('modules', 'detectors').value()['all_items']
        if detectors != self.modules_manager.detectors_name:
            messagebox(text='The loaded detectors from the batch file do not corresponds to the dashboard detectors')
            return

        self.settings = settings_tmp

        # for child in children:
        #     self.add_scan(name=child.name(), title=child.opts['title'])
        #
        #     self.settings.child('scans', child.name()).restoreState(child.saveState())

        if show:
            status = self.show_tree()
        else:
            self.settings_tree.setParameters(self.settings, showTop=False)
        return status

    def set_scans(self):
        infos = []
        acts, dets = self.get_act_dets()
        self._scans = OrderedDict([])
        for name in [child.name() for child in self.settings.child('scans').children()]:
            self._scans[name] = Scanner(actuators=self.modules_manager.get_mods_from_names(acts[name], 'act'))
            self._scans[name].set_scan_from_settings(self.settings.child('scans', name, Scanner.settings_name),
                                                     self.settings.child('scans', name, ScannerBase.settings_name))
            infos.append(f'{name}: {acts[name]} / {dets[name]}')
            infos.append(f'{name}: {self._scans[name].get_scan_info()}')
        return infos

    def get_scan(self, name: str):
        """Get a Scanner object from name"""
        if len(self._scans) == 0:
            self.set_scans()
        return self._scans.get(name)

    @property
    def scans(self):
        return self._scans

    def get_scan_names(self) -> List[str]:
        return list(self._scans.keys())

    def set_new_batch(self):
        self.settings = self.create_parameter(self.params)
        status = self.show_tree()
        return status

    def show_tree(self):


        dialog = QtWidgets.QDialog()
        dialog.setLayout(QtWidgets.QVBoxLayout())

        widget_all_settings = QtWidgets.QWidget()

        dialog.layout().addWidget(widget_all_settings)
        buttonBox = QtWidgets.QDialogButtonBox(parent=dialog)
        buttonBox.addButton('Save', buttonBox.AcceptRole)
        buttonBox.accepted.connect(dialog.accept)
        buttonBox.addButton('Cancel', buttonBox.RejectRole)
        buttonBox.rejected.connect(dialog.reject)

        dialog.layout().addWidget(buttonBox)
        dialog.setWindowTitle('Fill in information about this Scan batch')

        widget_all_settings.setLayout(QtWidgets.QHBoxLayout())
        widget_all_settings.layout().addWidget(self.settings_tree)

        widget_vertical = QtWidgets.QWidget()
        widget_vertical.setLayout(QtWidgets.QVBoxLayout())
        widget_all_settings.layout().addWidget(widget_vertical)

        self.scanner_widget = self.scanner.parent_widget
        add_scan = QtWidgets.QPushButton('Add Scan')
        add_scan.clicked.connect(self.add_scan)

        widget_vertical.layout().addWidget(self.modules_manager.settings_tree)
        widget_vertical.layout().addWidget(self.scanner_widget)
        widget_vertical.layout().addWidget(add_scan)

        res = dialog.exec()

        if res == dialog.Accepted:
            ioxml.parameter_to_xml_file(
                self.settings, self.batch_path.joinpath(self.settings.child('filename').value()))

        return res == dialog.Accepted

    def set_scanner_settings(self, settings_tree: QtWidgets.QWidget):
        while True:
            child = self.scanner_widget.layout().takeAt(0)
            if not child:
                break
            child.widget().deleteLater()
            QtWidgets.QApplication.processEvents()

        self.scanner_widget.layout().addWidget(settings_tree)

    def update_actuators(self, actuators: List[str]):
        self.scanner.actuators = self.modules_manager.actuators

    def add_scan(self, name=None, title=None):
        if name is None or name is False:
            name_prefix = 'scan'
            child_indexes = [int(par.name()[len(name_prefix) + 1:]) for par in self.settings.child('scans').children()]
            if child_indexes == []:
                newindex = 0
            else:
                newindex = max(child_indexes) + 1
            name = f'{name_prefix}{newindex:02.0f}'
            title = f'Scan {newindex:02.0f}'

        child = {'title': title, 'name': name, 'type': 'group', 'removable': True, 'children': params}

        # self._scans[name] = Scanner(actuators=self.modules_manager.actuators)

        self.settings.child('scans').addChild(child)
        self.settings.child('scans', name, 'modules',
                            'actuators').setValue(dict(all_items=self.modules_manager.actuators_name,
                                                       selected=self.modules_manager.selected_actuators_name))
        self.settings.child('scans', name, 'modules',
                            'detectors').setValue(dict(all_items=self.modules_manager.detectors_name,
                                                       selected=self.modules_manager.selected_detectors_name))

        self.settings.child('scans', name).addChild(self.create_parameter(self.scanner.settings))
        self.settings.child('scans', name).addChild(self.create_parameter(self.scanner.get_scanner_sub_settings()))


class BatchScanner(QtCore.QObject):
    def __init__(self, area, actuators=[], detectors=[]):
        super().__init__()
        self.dockarea = area
        self.batchmanager = BatchManager(actuators=actuators, detectors=detectors, path=batch_path)

    @property
    def scans(self):
        return self.batchmanager.scans

    @property
    def scans_names(self):
        return self.batchmanager.get_scan_names()

    def get_scan(self, name: str):
        return self.batchmanager.get_scan(name)

    def get_act_dets(self):
        return self.batchmanager.get_act_dets()

    def setupUI(self):
        # %% create scan dock and make it a floating window
        self.batch_dock = Dock("BatchScanner", size=(1, 1), autoOrientation=False)  # give this dock the minimum possible size
        self.dockarea.addDock(self.batch_dock, 'left')
        self.batch_dock.float()

        self.widget = QtWidgets.QWidget()
        self.widget.setLayout(QtWidgets.QVBoxLayout())

        widget_infos = QtWidgets.QWidget()
        self.widget_infos_list = QtWidgets.QListWidget()
        widget_infos.setLayout(QtWidgets.QHBoxLayout())
        splitter = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
        widget_infos.layout().addWidget(splitter)
        splitter.addWidget(self.batchmanager.settings_tree)
        splitter.addWidget(self.widget_infos_list)
        self.batch_dock.addWidget(self.widget)
        self.widget.layout().addWidget(widget_infos)
        #self.create_menu()

    def create_menu(self, menubar=None):
        """
        """
        # %% create Settings menu
        if menubar is None:
            menubar = QtWidgets.QMenuBar()
            self.widget.layout().insertWidget(0, menubar)

        self.batch_menu = menubar.addMenu('Batch Configs')
        action_new = self.batch_menu.addAction('New Batch')
        action_new.triggered.connect(self.batchmanager.set_new_batch)
        action_modify = self.batch_menu.addAction('Modify Batch')
        action_modify.triggered.connect(self.batchmanager.set_file_batch)
        self.batch_menu.addSeparator()
        self.load_batch = self.batch_menu.addMenu('Load Batchs')

        slots = dict([])
        for ind_file, file in enumerate(batch_path.iterdir()):
            if file.suffix == '.xml':
                filestem = file.stem
                slots[filestem] = self.load_batch.addAction(filestem)
                slots[filestem].triggered.connect(
                    self.create_menu_slot(batch_path.joinpath(file)))

    def load_file(self, filepath=None):
        if filepath is None:
            path = pymodaq.utils.gui_utils.file_io.select_file(start_path=batch_path, save=False, ext='xml')
            if path != '':
                filepath = path
            else:
                return
        self.batchmanager.set_file_batch(str(filepath), show=False)

        infos = self.batchmanager.set_scans()
        self.widget_infos_list.addItems(infos)

    def create_menu_slot(self, filename):
        return lambda: self.load_file(filename)


def main_batch_scanner():
    from pymodaq.control_modules.mocks import MockDAQMove, MockDAQViewer
    app = QtWidgets.QApplication(sys.argv)
    win = QtWidgets.QMainWindow()
    area = pymodaq.utils.gui_utils.dock.DockArea()
    win.setCentralWidget(area)

    # prog = BatchManager(msgbox=False, actuators=['Xaxis', 'Yaxis'], detectors=['Det0D', 'Det1D'])
    # prog.set_file_batch('C:\\Users\\weber\\pymodaq_local\\batch_configs\\batch_default.xml', show=False)
    # prog.set_scans()
    actuators = [MockDAQMove(title='Xaxis'), MockDAQMove(title='Yaxis')]
    detectors = [MockDAQViewer(title='Det0D'), MockDAQViewer(title='Det1D')]
    main = BatchScanner(area, actuators=actuators, detectors=detectors)
    main.setupUI()
    main.create_menu()
    win.show()
    sys.exit(app.exec_())


def main_batch_manager():
    from pymodaq.control_modules.mocks import MockDAQMove, MockDAQViewer
    app = QtWidgets.QApplication(sys.argv)
    actuators = [MockDAQMove(title='Xaxis'), MockDAQMove(title='Yaxis')]
    detectors = [MockDAQViewer(title='Det0D'), MockDAQViewer(title='Det1D')]
    prog = BatchManager(msgbox=True, actuators=actuators, detectors=detectors)
    sys.exit(app.exec_())


if __name__ == '__main__':
    #main_batch_manager()
    main_batch_scanner()



# -*- coding: utf-8 -*-
"""
Created the 04/11/2022

@author: Sebastien Weber
"""

import pyqtgraph as pg
from qtpy import QtWidgets

from pymodaq.utils.plotting.utils.plot_utils import View_cust
from pymodaq.utils.plotting.items.axis_scaled import AXIS_POSITIONS, AxisItem_Scaled


class ImageWidget(pg.GraphicsLayoutWidget):
    """this gives a layout to add imageitems.
    """

    def __init__(self, parent=None, *args_plotitem, **kwargs_plotitem):

        super().__init__(parent)
        self.setupUI(*args_plotitem, **kwargs_plotitem)

    def setAspectLocked(self, lock=True, ratio=1):
        """
        Defines the aspect ratio of the view
        Parameters
        ----------
        lock: (bool) if True aspect ratio is set to ratio, else the aspect ratio is varying when scaling the view
        ratio: (int) aspect ratio between horizontal and vertical axis
        """
        self.plotitem.vb.setAspectLocked(lock=True, ratio=1)

    def getAxis(self, position):
        return self.plotitem.getAxis(position)

    def setupUI(self, *args_plotitem, **kwargs_plotitem):
        layout = QtWidgets.QGridLayout()
        # set viewer area
        self.scene_obj = self.scene()
        self.view = View_cust()
        self.plotitem = pg.PlotItem(viewBox=self.view, *args_plotitem, **kwargs_plotitem)
        self.plotItem = self.plotitem  # for backcompatibility
        self.setAspectLocked(lock=True, ratio=1)
        self.setCentralItem(self.plotitem)

    def add_scaled_axis(self, position):
        """
        Add a AxisItem_Scaled to the given position with respect with the plotitem
        Parameters
        ----------
        position: (str) either 'top', 'bottom', 'right' or 'left'

        Returns
        -------

        """
        if position not in AXIS_POSITIONS:
            raise ValueError(f'The Axis position {position} should be in {AXIS_POSITIONS}')
        axis = AxisItem_Scaled(position)
        self.plotitem.setAxisItems({position: axis})
        return axis


class PlotWidget(pg.PlotWidget):
    def __init__(self, *args, **kwargs):
        plot_item = pg.PlotItem(viewBox=View_cust())
        super().__init__(*args, plotItem=plot_item, **kwargs)

    @property
    def view(self):
        return self.getViewBox()

    @property
    def legend(self):
        return self.plotItem.legend



# -*- coding: utf-8 -*-
"""
Created the 03/11/2022

@author: Sebastien Weber
"""
from collections import OrderedDict
from pyqtgraph import PlotCurveItem
from qtpy.QtCore import QObject, Signal, Slot
from pymodaq.utils.plotting.utils.plot_utils import Data0DWithHistory
from pymodaq.utils.managers.roi_manager import ROIManager
from pymodaq.utils.plotting.items.crosshair import Crosshair
from pymodaq.utils.logger import set_logger, get_module_name
import pymodaq.utils.daq_utils as utils
from pymodaq.utils.logger import set_logger, get_module_name


logger = set_logger(get_module_name(__file__))
IMAGE_TYPES = ['red', 'green', 'blue']
COLOR_LIST = utils.plot_colors
COLORS_DICT = dict(red=(255, 0, 0), green=(0, 255, 0), blue=(0, 0, 255), spread=(128, 128, 128))


def curve_item_factory(pen='red'):
    """
    Create a PlotCurveItem with the given pen
    Parameters
    ----------
    pen: any type of arguments accepted by pyqtgraph.function.mkColor or one of the COLORS_DICT key

    Returns
    -------
    PlotCurveItem
    """
    if isinstance(pen, str):
        if pen in COLORS_DICT:
            pen = COLORS_DICT[pen]
    return PlotCurveItem(pen=pen)


class LineoutPlotter(QObject):
    """Base class to manage and display data filtered out into lineouts (1D, 0D)

    Should be inherited and subclass some methods as appropriate

    Parameters
    ----------
    graph_widgets: OrderedDict
        Includes plotwidgets to display data
    roi_manager:
        The ROIManager to create ROIs and manage their properties
    crosshair:
        The Crosshair object
    """

    roi_changed = Signal(dict)
    crosshair_lineout_plotted = Signal(dict)
    roi_lineout_plotted = Signal(dict)

    lineout_widgets = ['int']  # should be reimplemented see viewer2D

    def __init__(self, graph_widgets: OrderedDict, roi_manager: ROIManager, crosshair: Crosshair):
        super().__init__()

        self._roi_manager = roi_manager
        self._crosshair = crosshair

        self._lineout_widgets = graph_widgets['lineouts']

        self.integrated_data = Data0DWithHistory()

        self._roi_curves = OrderedDict()
        self._crosshair_curves = OrderedDict()
        self._data_integrated = []

        self.setup_crosshair()

        self._roi_manager.new_ROI_signal.connect(self.add_ROI)
        self._roi_manager.remove_ROI_signal.connect(self.remove_ROI)
        self._roi_manager.roi_value_changed.connect(self.update_roi)

    def plot_roi_lineouts(self, roi_dicts):
        self.integrated_data.add_datas({roi_key: roi_dicts[roi_key].int_data for roi_key in roi_dicts})
        for roi_key, lineout_data in roi_dicts.items():
            if roi_key in self._roi_curves:
                self._roi_curves[roi_key]['int'].setData(self.integrated_data.xaxis,
                                                         self.integrated_data.datas[roi_key])
        self.plot_other_lineouts(roi_dicts)

        logger.debug('roi lineouts plotted')
        self.roi_lineout_plotted.emit(roi_dicts)

    def plot_other_lineouts(self, roi_dicts):
        raise NotImplementedError

    def plot_crosshair_lineouts(self, crosshair_dict):
        self.plot_other_crosshair_lineouts(crosshair_dict)

        logger.debug('crosshair lineouts plotted')
        self.crosshair_lineout_plotted.emit(crosshair_dict)

    def plot_other_crosshair_lineouts(self, crosshair_dict):
        raise NotImplementedError

    def get_lineout_widget(self, name):
        if name not in self.lineout_widgets:
            raise KeyError(f'The lineout_widget reference should be within {self.lineout_widgets} not {name}')
        return self._lineout_widgets[name]

    @Slot(str, tuple)
    def update_roi(self, roi_key, param_changed):
        param, param_value = param_changed

        if param.name() == 'Color':
            for curve in self._roi_curves[roi_key].values():
                curve.setPen(param_value)

        self.roi_changed.emit(self._roi_manager.ROIs)

    @Slot(str)
    def remove_ROI(self, roi_name):
        index = int(roi_name.split('_')[1])
        self.remove_roi_lineout_items(index)
        self.roi_changed.emit(self._roi_manager.ROIs)

    @Slot(int, str)
    def add_ROI(self, newindex, roi_type):
        item = self._roi_manager.get_roi_from_index(newindex)
        item.sigRegionChangeFinished.connect(lambda: self.roi_changed.emit(self._roi_manager.ROIs))
        item_param = self._roi_manager.settings.child('ROIs', 'ROI_{:02d}'.format(newindex))
        color = item_param.child('Color').value()

        self.add_roi_lineout_items(newindex, color)
        self.roi_changed.emit(self._roi_manager.ROIs)

    def add_roi_lineout_items(self, index, pen):
        """
        Add specifics lineouts generated from ROIs
        Parameters
        ----------
        index: (int) index of the ROI generating these lineouts
        pen: (str, tuple) any argument able to generate a QPen, see pyqtgraph.functions.mkPen
        """
        self._roi_curves[f'ROI_{index:02d}'] = \
            {curv_key: curve_item_factory(pen) for curv_key in self.lineout_widgets}
        self.add_lineout_items(*self._roi_curves[f'ROI_{index:02d}'].values())

    def remove_roi_lineout_items(self, index):
        """
        Remove specifics lineouts generated from ROI referenced by a unique integer
        Parameters
        ----------
        index: (int) index of the ROI generating these lineouts
        """
        items = self._roi_curves.pop(f'ROI_{index:02d}')
        self.remove_lineout_items(*items.values())

    def add_lineout_items(self, *curve_items):
        """
        Add Curve items sequentially to lineouts widgets: (hor, ver and int)
        Parameters
        ----------
        curve_items: (PlotCurveItem) at most 3 of them
        """
        for ind, curve_item in enumerate(curve_items):
            self.get_lineout_widget(self.lineout_widgets[ind]).addItem(curve_item)

    def remove_lineout_items(self, *curve_items):
        """
        Remove Curve items sequentially to lineouts widgets: (hor, ver and int)
        Parameters
        ----------
        curve_items: (PlotCurveItem) at most 3 of them
        """

        for ind, curve_item in enumerate(curve_items):
            self.get_lineout_widget(self.lineout_widgets[ind]).removeItem(curve_item)

    @Slot(bool)
    def roi_clicked(self, isroichecked=True):
        self._roi_manager.roiwidget.setVisible(isroichecked)

        for k, roi in self._roi_manager.ROIs.items():
            roi.setVisible(isroichecked)
            for item in self.get_roi_curves_triplet()[k].values():
                item.setVisible(isroichecked)

    @Slot(bool)
    def crosshair_clicked(self, iscrosshairchecked=True):
        for image_key in IMAGE_TYPES:
            self.show_crosshair_curves(image_key, iscrosshairchecked)

    def get_roi_curves_triplet(self):
        """
        Get the dictionary (one key by ROI) containing dicts with ROI PlotCurveItem

        Example:
        --------
        >>> roi_dict_triplet = self.get_roi_cruves_triplet()
        >>> hor_curve = roi_dict_triplet['ROI_00']['hor']  # where 'hor' is an entry of self.lineout_widgets
        """
        return self._roi_curves

    def get_crosshair_curves_triplet(self):
        """
        Get the dictionary (one key by ImageItem, see IMAGE_TYPES) containing dicts with PlotCurveItem

        Example:
        --------
        >>> crosshair_dict_triplet = self.get_crosshair_curves_triplet()
        >>> hor_curve = crosshair_dict_triplet['blue']['hor']  # where 'hor' is an entry of self.lineout_widgets
        """
        return self._crosshair_curves

    def get_crosshair_curve_triplet(self, curve_name):
        return self._crosshair_curves[curve_name]

    def setup_crosshair(self):
        """to reimplement if needed"""
        pass

    def show_crosshair_curves(self, curve_key, show=True):
        """to reimplement if needed"""
        pass



# -*- coding: utf-8 -*-
"""
Created the 25/10/2022

@author: Sebastien Weber
"""
from .utils import get_extensions
from .pid.utils import get_models

from .console import QtConsole
from .daq_scan import DAQScan
from .daq_logger import DAQ_Logger
from .pid.pid_controller import DAQ_PID
from .h5browser import H5Browser





from qtpy import QtCore, QtGui, QtWidgets


class Image_Viewer(QtWidgets.QGraphicsView):
    def __init__(self, parent):
        super(Image_Viewer, self).__init__(parent)
        self._zoom = 0
        self._scene = QtWidgets.QGraphicsScene(self)
        self._image = QtWidgets.QGraphicsPixmapItem()
        self._scene.addItem(self._image)
        self.setScene(self._scene)
        self.setTransformationAnchor(QtWidgets.QGraphicsView.AnchorUnderMouse)
        self.setResizeAnchor(QtWidgets.QGraphicsView.AnchorUnderMouse)

        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setBackgroundBrush(QtGui.QBrush(QtGui.QColor(30, 30, 30)))
        self.setFrameShape(QtWidgets.QFrame.NoFrame)

    def fitInView(self):
        rect = QtCore.QRectF(self._image.pixmap().rect())
        if not rect.isNull():
            unity = self.transform().mapRect(QtCore.QRectF(0, 0, 1, 1))
            self.scale(1 / unity.width(), 1 / unity.height())
            viewrect = self.viewport().rect()
            scenerect = self.transform().mapRect(rect)
            factor = min(viewrect.width() / scenerect.width(),
                         viewrect.height() / scenerect.height())
            self.scale(factor, factor)
            self.centerOn(rect.center())
            self._zoom = 0

    def setPhoto(self, pixmap=None):
        self._zoom = 0
        if pixmap and not pixmap.isNull():
            self.setDragMode(QtWidgets.QGraphicsView.ScrollHandDrag)
            self._image.setPixmap(pixmap)
            # self.fitInView()
        else:
            self.setDragMode(QtWidgets.QGraphicsView.NoDrag)
            self._image.setPixmap(QtGui.QPixmap())

    def zoomFactor(self):
        return self._zoom

    def wheelEvent(self, event):
        try:
            if not self._image.pixmap().isNull():
                # delta=QtCore.QPoint()
                # delta.y
                if event.angleDelta().y() > 0:
                    factor = 1.25
                    self._zoom += 1
                else:
                    factor = 0.8
                    self._zoom -= 1
                if self._zoom > 0:
                    self.scale(factor, factor)
                elif self._zoom == 0:
                    self.fitInView()
                else:
                    self._zoom = 0
        except Exception as e:
            pass


class Window(QtWidgets.QWidget):
    def __init__(self):
        super(Window, self).__init__()
        self.viewer = Image_Viewer(self)
        self.edit = QtWidgets.QLineEdit(self)
        self.edit.setReadOnly(True)
        self.button = QtWidgets.QToolButton(self)
        self.button.setText('...')
        self.button.clicked.connect(self.handleOpen)
        layout = QtWidgets.QGridLayout(self)
        layout.addWidget(self.viewer, 0, 0, 1, 2)
        layout.addWidget(self.edit, 1, 0, 1, 1)
        layout.addWidget(self.button, 1, 1, 1, 1)

    def handleOpen(self):
        path = QtWidgets.QFileDialog.getOpenFileName(
            self, 'Choose Image', self.edit.text())
        path = path[0]
        if path:
            self.edit.setText(path)
            self.viewer.setPhoto(QtGui.QPixmap(path))


if __name__ == '__main__':
    import sys

    app = QtWidgets.QApplication(sys.argv)
    window = Window()
    window.setGeometry(500, 300, 800, 600)
    window.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'DAQ_Measurement_GUI.ui'
#
# Created by: qtpy UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from qtpy import QtCore, QtGui, QtWidgets
from pymodaq.resources.QtDesigner_Ressources import QtDesigner_ressources_rc


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(754, 326)
        self.gridLayout_2 = QtWidgets.QGridLayout(Form)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.splitter_2 = QtWidgets.QSplitter(Form)
        self.splitter_2.setOrientation(QtCore.Qt.Horizontal)
        self.splitter_2.setObjectName("splitter_2")
        self.splitter = QtWidgets.QSplitter(self.splitter_2)
        self.splitter.setOrientation(QtCore.Qt.Vertical)
        self.splitter.setObjectName("splitter")
        self.gridLayoutWidget = QtWidgets.QWidget(self.splitter)
        self.gridLayoutWidget.setObjectName("gridLayoutWidget")
        self.gridLayout = QtWidgets.QGridLayout(self.gridLayoutWidget)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setObjectName("gridLayout")
        self.formula_edit = QtWidgets.QPlainTextEdit(self.gridLayoutWidget)
        self.formula_edit.setMaximumSize(QtCore.QSize(16777215, 50))
        self.formula_edit.setPlainText("")
        self.formula_edit.setTextInteractionFlags(QtCore.Qt.TextSelectableByKeyboard | QtCore.Qt.TextSelectableByMouse)
        self.formula_edit.setObjectName("formula_edit")
        self.gridLayout.addWidget(self.formula_edit, 4, 1, 1, 1)
        self.Quit_pb = QtWidgets.QPushButton(self.gridLayoutWidget)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/Icon_Library/close2.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.Quit_pb.setIcon(icon)
        self.Quit_pb.setObjectName("Quit_pb")
        self.gridLayout.addWidget(self.Quit_pb, 1, 0, 1, 2)
        self.measurement_type_combo = QtWidgets.QComboBox(self.gridLayoutWidget)
        self.measurement_type_combo.setObjectName("measurement_type_combo")
        self.measurement_type_combo.addItem("")
        self.measurement_type_combo.addItem("")
        self.measurement_type_combo.addItem("")
        self.measurement_type_combo.addItem("")
        self.measurement_type_combo.addItem("")
        self.measurement_type_combo.addItem("")
        self.gridLayout.addWidget(self.measurement_type_combo, 2, 1, 1, 1)
        self.label_11 = QtWidgets.QLabel(self.gridLayoutWidget)
        self.label_11.setAlignment(QtCore.Qt.AlignCenter)
        self.label_11.setObjectName("label_11")
        self.gridLayout.addWidget(self.label_11, 3, 0, 1, 1)
        self.label = QtWidgets.QLabel(self.gridLayoutWidget)
        font = QtGui.QFont()
        font.setFamily("Tahoma")
        font.setPointSize(14)
        font.setBold(True)
        font.setItalic(True)
        font.setWeight(75)
        self.label.setFont(font)
        self.label.setAlignment(QtCore.Qt.AlignCenter)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 0, 0, 1, 2)
        self.measure_subtype_combo = QtWidgets.QComboBox(self.gridLayoutWidget)
        self.measure_subtype_combo.setObjectName("measure_subtype_combo")
        self.measure_subtype_combo.addItem("")
        self.measure_subtype_combo.addItem("")
        self.measure_subtype_combo.addItem("")
        self.gridLayout.addWidget(self.measure_subtype_combo, 3, 1, 1, 1)
        self.label_9 = QtWidgets.QLabel(self.gridLayoutWidget)
        self.label_9.setAlignment(QtCore.Qt.AlignCenter)
        self.label_9.setObjectName("label_9")
        self.gridLayout.addWidget(self.label_9, 2, 0, 1, 1)
        self.label_7 = QtWidgets.QLabel(self.gridLayoutWidget)
        self.label_7.setAlignment(QtCore.Qt.AlignCenter)
        self.label_7.setObjectName("label_7")
        self.gridLayout.addWidget(self.label_7, 4, 0, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem, 2, 2, 1, 1)
        self.label_4 = QtWidgets.QLabel(self.gridLayoutWidget)
        self.label_4.setObjectName("label_4")
        self.gridLayout.addWidget(self.label_4, 5, 0, 1, 1)
        self.result_sb = QtWidgets.QDoubleSpinBox(self.gridLayoutWidget)
        font = QtGui.QFont()
        font.setPointSize(8)
        self.result_sb.setFont(font)
        self.result_sb.setReadOnly(True)
        self.result_sb.setButtonSymbols(QtWidgets.QAbstractSpinBox.NoButtons)
        self.result_sb.setMinimum(-100000000.0)
        self.result_sb.setMaximum(100000000.0)
        self.result_sb.setObjectName("result_sb")
        self.gridLayout.addWidget(self.result_sb, 5, 1, 1, 1)
        self.horizontalLayoutWidget = QtWidgets.QWidget(self.splitter)
        self.horizontalLayoutWidget.setObjectName("horizontalLayoutWidget")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.horizontalLayoutWidget)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.result_lcd = QtWidgets.QLCDNumber(self.horizontalLayoutWidget)
        self.result_lcd.setSmallDecimalPoint(True)
        self.result_lcd.setDigitCount(10)
        self.result_lcd.setObjectName("result_lcd")
        self.horizontalLayout.addWidget(self.result_lcd)
        self.gridLayout_2.addWidget(self.splitter_2, 0, 0, 1, 1)
        self.StatusBarLayout = QtWidgets.QHBoxLayout()
        self.StatusBarLayout.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.StatusBarLayout.setObjectName("StatusBarLayout")
        self.gridLayout_2.addLayout(self.StatusBarLayout, 1, 0, 1, 1)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.Quit_pb.setText(_translate("Form", "Quit"))
        self.measurement_type_combo.setItemText(0, _translate("Form", "Cursor Intensity Integration"))
        self.measurement_type_combo.setItemText(1, _translate("Form", "Max"))
        self.measurement_type_combo.setItemText(2, _translate("Form", "Min"))
        self.measurement_type_combo.setItemText(3, _translate("Form", "Gaussian Fit"))
        self.measurement_type_combo.setItemText(4, _translate("Form", "Lorentzian Fit"))
        self.measurement_type_combo.setItemText(5, _translate("Form", "Exponential Decay Fit"))
        self.label_11.setText(_translate("Form", "Measurement SubType: "))
        self.label.setText(_translate("Form", "DAQ Measurement"))
        self.measure_subtype_combo.setItemText(0, _translate("Form", "dx"))
        self.measure_subtype_combo.setItemText(1, _translate("Form", "amplitude"))
        self.measure_subtype_combo.setItemText(2, _translate("Form", "x0"))
        self.label_9.setText(_translate("Form", "Measurement Type: "))
        self.label_7.setText(_translate("Form", "Formula\n"
                                                "(the variable is \"x\"):"))
        self.label_4.setText(_translate("Form", "Result:"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    Form = QtWidgets.QWidget()
    ui = Ui_Form()
    ui.setupUi(Form)
    Form.show()
    sys.exit(app.exec_())
